"""
Special symbols that can be used in kernels and contain information about the current block
"""
import sympy as sp
from sympy.core.cache import cacheit

import pystencils
from pystencils.data_types import TypedSymbol, create_type


class AABBMax(TypedSymbol):
    """
    Local cell interval in global index coordinates
    """
    def __new__(cls, *args, **kwds):
        obj = AABBMax.__xnew_cached_(cls, *args, **kwds)
        return obj

    def __new_stage2__(cls, dim, *args, **kwargs):
        obj = super(AABBMax, cls).__xnew__(cls,
                                           f'_aabb_max_{dim}',
                                           create_type('double'),
                                           *args,
                                           **kwargs)
        obj.dim = dim
        return obj

    __xnew__ = staticmethod(__new_stage2__)
    __xnew_cached_ = staticmethod(cacheit(__new_stage2__))

    def _hashable_content(self):
        return super()._hashable_content(), self.dim, self.__class__.__name__

    def __getnewargs__(self):
        return self.dim


class AABBMin(TypedSymbol):
    """
    Local cell interval in global index coordinates
    """
    def __new__(cls, *args, **kwds):
        obj = AABBMin.__xnew_cached_(cls, *args, **kwds)
        return obj

    def __new_stage2__(cls, dim, *args, **kwargs):
        obj = super(AABBMin, cls).__xnew__(cls,
                                           f'_aabb_min_{dim}',
                                           create_type('double'),
                                           *args,
                                           **kwargs)
        obj.dim = dim
        return obj

    __xnew__ = staticmethod(__new_stage2__)
    __xnew_cached_ = staticmethod(cacheit(__new_stage2__))

    def _hashable_content(self):
        return super()._hashable_content(), self.dim, self.__class__.__name__

    def __getnewargs__(self):
        return self.dim


class Dx(TypedSymbol):
    """
    Local cell interval in global index coordinates
    """
    def __new__(cls, *args, **kwds):
        obj = Dx.__xnew_cached_(cls, *args, **kwds)
        return obj

    def __new_stage2__(cls, dim, *args, **kwargs):
        obj = super(Dx, cls).__xnew__(cls,
                                      f'_dx_{dim}',
                                      create_type('double'),
                                      *args,
                                      **kwargs)
        obj.dim = dim
        return obj

    __xnew__ = staticmethod(__new_stage2__)
    __xnew_cached_ = staticmethod(cacheit(__new_stage2__))

    def _hashable_content(self):
        return super()._hashable_content(), self.dim, self.__class__.__name__

    def __getnewargs__(self):
        return self.dim


dx_vector = sp.Matrix([Dx(i) for i in range(3)])
dx = Dx(0)
dy = Dx(1)
dz = Dx(2)
aabb_min_vector = sp.Matrix([AABBMin(i) for i in range(3)])
aabb_min_x = AABBMin(0)
aabb_min_y = AABBMin(1)
aabb_min_z = AABBMin(2)
aabb_max_vector = sp.Matrix([AABBMax(i) for i in range(3)])
aabb_max_x = AABBMax(0)
aabb_max_y = AABBMax(1)
aabb_max_z = AABBMax(2)
# TODO this assumes excactly one ghost layer
global_coord = aabb_min_vector + sp.Matrix([i*(j-1) for i, j in zip(dx_vector, pystencils.x_vector(3))])
global_coord_x = global_coord[0]
global_coord_y = global_coord[1]
global_coord_z = global_coord[2]
